#include <cstdio>
#include <cstring>
#include <cmath>

using namespace std;

#include "snd.hh"
#include "instrument.hh"
#include "sndplayer.hh"
#include "analysis.hh"

static void DumpSND(const song &song)
{
    for(unsigned a=0; a<repository.size(); ++a)
    {
        fprintf(stderr, "Instrument %u:\n", a+1);
        DumpSample(repository[a]);
    }

    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const SNDChannel &tmp = song.ChanPtr(a, c);
            char Buf[64];
            int colormode=0;
            
            sprintf(Buf, "v%Xs%dw%ua%u", tmp.vol, tmp.sample, tmp.wavelen, tmp.age);
            if(tmp.ins)
                sprintf(strchr(Buf,0), "(%u)", tmp.ins);
            
            sprintf(Buf+32, "%18s", Buf);
            if(tmp.vol)
            {
                if(IsBeat(a, c))
                {
                    fputs("\33[1;34m", stderr);
                    colormode=2;
                }
                else
                    colormode=1;
            }
            else
            {
                colormode=0;
                fputs("\33[1;30m", stderr);
            }
            for(char *s=Buf+32; *s; ++s)
            {
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[1;37m", stderr);
                */
                fputc(*s, stderr    );
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[m", stderr);
                */
            }
            fputs("\33[m", stderr);
        }
        fprintf(stderr, "\n");
    }
}

static unsigned NesVoltoS3Mvol(unsigned v)
{
    static unsigned char voltable[16] =
    {0,32,38,40,
     42,46,48,50,
     52,54,56,58,
     60,62,63,64};
    return voltable[v];
}

static void Patternize(const song &song)
{
    /* TODO: Handle the sample repository */
    /* TODO: Handle DPCM samples */
    
    /* Handle each channel separately */
    for(unsigned chan=0; chan<song.channelcount; ++chan)
    {
        unsigned nuotti=0, basewave=0;
        
        unsigned t=0;
        unsigned lastrow=0;
        bool wason = false;
        for(double pos=0; ; pos += TickSpacer)
        {
            unsigned row = (unsigned)pos;
            if(row >= song.rownum)break;
        
            const SNDChannel &tmp = song.ChanPtr(row, chan);
            
            // vol=tmp.vol     (volume(0..15))
            // smp=tmp.sample  (sqr:duty; pcm:pcm sample no)
            // wav=tmp.wavelen (wave length)
            // age=tmp.age     (age since last touched)
            // ins=tmp.ins     (index to repository)
            // beat=beat
            const SNDChannel *beatrow = NULL;
            bool beat = false;
            while(lastrow <= row)
                if(IsBeat(lastrow++, chan))
                {
                	beatrow = &song.ChanPtr(lastrow-1, chan);
                    beat = true;
                }
            
            fprintf(stderr, "%u ", t++%4);
            
            if(beat)
            {
                static const char nuotit[12*3+1] = "C-\0C#\0D-\0D#\0E-\0F-\0F#\0G-\0G#\0A-\0A#\0B-\0";
                basewave = tmp.wavelen;
                nuotti = (unsigned)(12 * log(4000.0 / basewave) / log(2.0));
                
                char Buf[64];
                sprintf(Buf, "%s%d", nuotit+3*(nuotti%12), nuotti/12);
                fprintf(stderr, "%s", Buf);
                wason = true;
            }
            else if(wason && !tmp.vol)
            {
                fprintf(stderr, "^^^");
                wason = false;
            }
            else
                fprintf(stderr, "...");
            
            if(tmp.type == SNDChannel::pcm && beat && tmp.sample)
                fprintf(stderr, " %02d", tmp.sample + repository.size());
            else if(tmp.type == SNDChannel::pcm && beat)
                fprintf(stderr, " %02d", beatrow->sample + repository.size());
            else if(tmp.ins)
                fprintf(stderr, " %02d", tmp.ins);
            else if(beatrow && beatrow->ins)
                fprintf(stderr, " %02d", beatrow->ins);
            else
                fprintf(stderr, " ..");
            
            if(beat)
            {
                int volavg = 0, volcount=0;
                for(unsigned tmprow = row; tmprow < row+8 && tmprow < song.rownum; ++tmprow)
                {
                    if(tmprow > row && IsBeat(tmprow, chan))break;
                    const SNDChannel &tmp = song.ChanPtr(tmprow, chan);
                    if(!tmp.vol)break;
                    volavg += tmp.vol; ++volcount;
                }
                volavg /= volcount;
                fprintf(stderr, " %02d", NesVoltoS3Mvol(volavg));
            }
            else
            {
                fprintf(stderr, " %02d", NesVoltoS3Mvol(tmp.vol));
            }
            
            fprintf(stderr, "\n");
        }
        fprintf(stderr, "----------\n");
    }
}

int main(int argc, const char *const *argv)
{
    if(argc != 2 || !strcmp(argv[1], "--help") || !strcmp(argv[1], "/?"))
    {
        puts("Simple SNDv3 player - Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n"
             "Usage:\n"
             "   snd <filename.snd>\n"
            );
        return 1;
    }
    
    song song = ReadSND(argv[1]);
    
    /* TODO: Handle the snd (dump it somewhere or something) */
    
    AnalyzeSND(song);
    DumpSND(song);
    Patternize(song);
    PlaySND(song);
    
    return 0;
}
